// Message to store the mapping from class label strings to class id. Datasets
// use string labels to represent classes while the object detection framework
// works with class ids. This message maps them so they can be converted back
// and forth as needed.
// Source modified from https://github.com/tensorflow/models/blob/master/research/object_detection/protos/string_int_label_map.proto
// Run `protoc --python_out=./ string_int_label_map.proto` under this directory
// to generate string_int_label_map_pb2.py file.
syntax = "proto2";

message StringIntLabelMapItem {
  // String name. The most common practice is to set this to a MID or synsets
  // id.
  optional string name = 1;

  // Integer id that maps to the string name above. Label ids should start from
  // 1.
  optional int32 id = 2;

  // Human readable string label.
  optional string display_name = 3;

  //
  optional string path = 4;
};

message StringIntLabelMap {
  repeated StringIntLabelMapItem item = 1;
};
